
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <malloc.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <chkstack.h>
#include <libsmbclient.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/times.h>

#include "../smbremote/smbc_api.h"
int smbc_redir_thread(void *pv);

pthread_key_t g_chkstack_key = -1;
pthread_mutex_t g_smbc_mutex;

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#define FIELD_OFFSET(type, field)  ((long)&(((type *)0)->field))
#define arrayof(x) (sizeof(x) / sizeof(*(x)))

#include "interruptible.c"

#if 0
extern void *_start, *_fini;
unsigned long checksum_smbclient(void)
{
    unsigned long v = 0;
    unsigned long *p;
    static unsigned long csum = 0;

    p = &_start;
    while (p < &_fini) {
	v += *p;
	p++;
    }
    if (!csum) {
	csum = v;
	printf("smbclient checksum is %lx\n", csum);
    }
    else if (v != csum) {
	printf("smbclient checksum mismatch %lx\n", v);
	sleep(1000);
    }
    return v;
}
#endif

#if defined(ARM)
long long get_tick_count(void)
{
    struct tms tms;
    return (times(&tms) * 1000) / sysconf(_SC_CLK_TCK);
}

__malloc_ptr_t my_morecore(ptrdiff_t __size)
{
    __malloc_ptr_t *p;
    static int oom = 0;
    static long long oom_time = 0;

    printf("[%08lx] MORECORE %x bytes\n", pthread_self(), __size);
    if (oom) {
	if (get_tick_count() < oom_time + 5000)
	    return (__malloc_ptr_t)-1;
	oom = 0;
    }
    p = __default_morecore(__size);
    if (p == (__malloc_ptr_t)-1) {
	printf("MORECORE FAILED %d bytes\n", __size);
	oom = TRUE;
	oom_time = get_tick_count();
    }
    return p;
}
#endif

void smbc_auth_fn(const char *server, const char *share,
		  char *workgroup, int wgmaxlen, char *username, int unmaxlen,
		  char *password, int pwmaxlen)
{   
    username[0] = password[0] = 0;
}

#ifdef SMBC_USE_SHARED_MEMORY
struct {
    unsigned char *base_address;
    int		   size;
    int		   id;
    unsigned long  owner;
} g_smbc_shmaps[10];

int smbc_alloc_shared(int size)
{
    int h;

    pthread_mutex_lock(&g_smbc_mutex);
    for (h = 0; h < arrayof(g_smbc_shmaps); h++) {
	if (!g_smbc_shmaps[h].base_address) {
	    g_smbc_shmaps[h].id = shmget('SMB0'+h, size, 0666 | IPC_CREAT);
	    if (g_smbc_shmaps[h].id < 0)
		break;
	    g_smbc_shmaps[h].base_address = shmat(g_smbc_shmaps[h].id, NULL, 0);
	    if (!g_smbc_shmaps[h].base_address) {
		shmctl(g_smbc_shmaps[h].id, IPC_RMID, 0);
		g_smbc_shmaps[h].id = -1;
		break;
	    }
	    g_smbc_shmaps[h].size = size;
	    g_smbc_shmaps[h].owner = pthread_self();
	    pthread_mutex_unlock(&g_smbc_mutex);
	    printf("smbclient: Allocated shared buffer of %d bytes as handle %d\n", size, h);
	    return h;
	}
    }
    printf("smbclient: Failed to allocate shared memory, %d bytes (h = %d)\n", size, h);
    pthread_mutex_unlock(&g_smbc_mutex);
    return -1;
}

int smbc_free_shared(int h)
{
    printf("smbclient: Freeing shared buffer %d\n", h);
    pthread_mutex_lock(&g_smbc_mutex);
    if (h >= 0 && h < arrayof(g_smbc_shmaps)) {
	if (g_smbc_shmaps[h].base_address) {
	    shmdt(g_smbc_shmaps[h].base_address);
	    g_smbc_shmaps[h].base_address = NULL;
	}
	if (g_smbc_shmaps[h].id >= 0) {
	    shmctl(g_smbc_shmaps[h].id, IPC_RMID, 0);
	    g_smbc_shmaps[h].id = -1;
	}
	g_smbc_shmaps[h].owner = 0;
    }
    pthread_mutex_unlock(&g_smbc_mutex);
    return 0;
}

void smbc_free_shared_owner(unsigned long owner)
{
    int h;

    pthread_mutex_lock(&g_smbc_mutex);
    for (h = 0; h < arrayof(g_smbc_shmaps); h++) {
	if (g_smbc_shmaps[h].base_address && owner == g_smbc_shmaps[h].owner) {
	    printf("smbclient: Cleaning up shared buffer %d\n", h);
	    shmdt(g_smbc_shmaps[h].base_address);
	    g_smbc_shmaps[h].base_address = NULL;
	    if (g_smbc_shmaps[h].id >= 0) {
		shmctl(g_smbc_shmaps[h].id, IPC_RMID, 0);
		g_smbc_shmaps[h].id = -1;
	    }
	    g_smbc_shmaps[h].owner = 0;
	}
    }
    pthread_mutex_unlock(&g_smbc_mutex);
}

void *smbc_unwire_ptr(int h, int i, int len)
{
    //printf("smbc_unwire_ptr, h = %d, i = %d, len = %d\n", h, i, len);

    if (h < 0 || h >= arrayof(g_smbc_shmaps) || !g_smbc_shmaps[h].base_address) {
	printf("smbclient: Invalid shared handle %d\n", h);
	return NULL;
    }
    if (i < 0 || i + len > g_smbc_shmaps[h].size) {
	printf("smbclient: Invalid shared offset/size %d-%d\n", i, i + len);
	return NULL;
    }

    return (char *)g_smbc_shmaps[h].base_address + i;
}
#endif

int smbc_dispatch_msg(SMBC_MSG *msg, SMBC_ACK_INFO *ack_info)
{
    int size;
    SMBC_OPEN_INFO *open_info;
    SMBC_READ_INFO *read_info;
    SMBC_WRITE_INFO *write_info;
    SMBC_LSEEK_INFO *lseek_info;
    SMBC_CLOSE_INFO *close_info;
    SMBC_UNLINK_INFO *unlink_info;
    SMBC_LOCK_INFO *lock_info;	
    SMBC_UNLOCK_INFO *unlock_info;
    SMBC_RENAME_INFO *rename_info;
    SMBC_OPENDIR_INFO *opendir_info;
    SMBC_CLOSEDIR_INFO *closedir_info;
    SMBC_READDIR_INFO *readdir_info;
    SMBC_TELLDIR_INFO *telldir_info;	
    SMBC_LSEEKDIR_INFO *lseekdir_info;
    SMBC_MKDIR_INFO *mkdir_info;
    SMBC_RMDIR_INFO *rmdir_info;
    SMBC_STAT_INFO *stat_info;
    SMBC_FSTAT_INFO *fstat_info;
    SMBC_CHOWN_INFO *chown_info;
    SMBC_CHMOD_INFO *chmod_info;
    SMBC_REMOVEUNUSEDSERVERS_INFO *rus_info;
    NMBC_NAMEQUERY_INFO *nq_info;
#ifdef SMBC_USE_SHARED_MEMORY
    SMBC_ALLOCSHARED_INFO *allocsh_info;
    SMBC_FREESHARED_INFO *freesh_info;
#endif
    struct smbc_dirent *dirent;

    CHKSTACK();

    reset_interruptible();

    errno = 0;
    switch (msg->type) {
    case SMBC_MSG_OPEN:
	open_info = msg;
#ifdef SMBC_USE_SHARED_MEMORY
	open_info->fname = smbc_unwire_ptr(open_info->h, open_info->fname, SMBC_MAX_PATH);
#endif
	//printf("smbc_message_loop: OPEN %s\n", open_info->fname);
	ack_info->value.iValue = smbc_open(open_info->fname, open_info->flags, open_info->mode);
	size = sizeof(int);
	break;
    case SMBC_MSG_READ:
	read_info = msg;
#ifdef SMBC_USE_SHARED_MEMORY
	read_info->buf = smbc_unwire_ptr(read_info->h, read_info->buf, read_info->count);
#endif
	//printf("smbc_message_loop: READ %d\n", read_info->count);
	ack_info->value.iValue = smbc_read(read_info->fd, read_info->buf, read_info->count);
	//printf("read %d bytes\n", ack_info->value.iValue);
	size = sizeof(int);
	break;
    case SMBC_MSG_WRITE:
	write_info = msg;
#ifdef SMBC_USE_SHARED_MEMORY
	write_info->buf = smbc_unwire_ptr(write_info->h, write_info->buf, write_info->count);
#endif
	//printf("smbc_message_loop: WRITE %d\n", write_info->count);
	ack_info->value.iValue = smbc_write(write_info->fd, write_info->buf, write_info->count);
	//printf("wrote %d bytes\n", ack_info->value.iValue);
	size = sizeof(int);
	break;
    case SMBC_MSG_LSEEK:
	lseek_info = msg;
	//printf("smbc_message_loop: LSEEK %d\n", lseek_info->fd);
	ack_info->value.llValue = smbc_lseek(lseek_info->fd, lseek_info->offset, lseek_info->whence);
	size = sizeof(long long);
	break;
    case SMBC_MSG_CLOSE:
	close_info = msg;
	//printf("smbc_message_loop: CLOSE %d\n", close_info->fd);
	ack_info->value.iValue = smbc_close(close_info->fd);
	size = sizeof(int);
	break;
    case SMBC_MSG_UNLINK:
	unlink_info = msg;
#ifdef SMBC_USE_SHARED_MEMORY
	unlink_info->furl = smbc_unwire_ptr(unlink_info->h, unlink_info->furl, SMBC_MAX_PATH);
#endif
	//printf("smbc_message_loop: UNLINK %s\n", unlink_info->furl);
	ack_info->value.iValue = smbc_unlink(unlink_info->furl);
	size = sizeof(int);
	break;
    case SMBC_MSG_LOCK:
	lock_info = msg;
	//printf("smbc_message_loop: LOCK %d\n", lock_info->fd);
	ack_info->value.iValue = smbc_lock(lock_info->fd, lock_info->offset, lock_info->len, lock_info->timeout, lock_info->lock_type);
	size = sizeof(int);
	break;
    case SMBC_MSG_UNLOCK:
	unlock_info = msg;
	//printf("smbc_message_loop: UNLOCK %d\n", unlock_info->fd);
	ack_info->value.iValue = smbc_unlock(unlock_info->fd, unlock_info->offset, unlock_info->len);
	size = sizeof(int);
	break;
    case SMBC_MSG_RENAME:
	rename_info = msg;
#ifdef SMBC_USE_SHARED_MEMORY
	rename_info->ourl = smbc_unwire_ptr(rename_info->h, rename_info->ourl, SMBC_MAX_PATH);
	rename_info->nurl = smbc_unwire_ptr(rename_info->h, rename_info->nurl, SMBC_MAX_PATH);
#endif
	//printf("smbc_message_loop: RENAME %s -> %s\n", rename_info->ourl, rename_info->nurl);
	ack_info->value.iValue = smbc_rename(rename_info->ourl, rename_info->nurl);
	size = sizeof(int);
	break;
    case SMBC_MSG_OPENDIR:
	opendir_info = msg;
#ifdef SMBC_USE_SHARED_MEMORY
	opendir_info->fname = smbc_unwire_ptr(opendir_info->h, opendir_info->fname, SMBC_MAX_PATH);
#endif
	//printf("smbc_message_loop: OPENDIR %s\n", opendir_info->fname);
	ack_info->value.iValue = smbc_opendir(opendir_info->fname);
	size = sizeof(int);
	break;
    case SMBC_MSG_CLOSEDIR:
	closedir_info = msg;
	//printf("smbc_message_loop: CLOSEDIR %d\n", closedir_info->fd);
	ack_info->value.iValue = smbc_closedir(closedir_info->fd);
	size = sizeof(int);
	break;
    case SMBC_MSG_READDIR:
	readdir_info = msg;
	//printf("smbc_message_loop: READDIR %d\n", readdir_info->fd);
	dirent = smbc_readdir(readdir_info->fd);
#ifdef SMBC_USE_SHARED_MEMORY
	if (dirent) {
	    ack_info->value.pvValue = smbc_unwire_ptr(readdir_info->h, 0, dirent->dirlen);
	    if (ack_info->value.pvValue)
		memcpy(ack_info->value.pvValue, dirent, dirent->dirlen);
	}
#else
	ack_info->value.pvValue = dirent;
#endif
	size = sizeof(void *);
	break;
    case SMBC_MSG_TELLDIR:
	telldir_info = msg;
	//printf("smbc_message_loop: TELLDIR %d\n", telldir_info->dh);
	ack_info->value.llValue = smbc_telldir(telldir_info->dh);
	size = sizeof(long long);
	break;
    case SMBC_MSG_LSEEKDIR:
	lseekdir_info = msg;
	//printf("smbc_message_loop: LSEEKDIR %d\n", lseekdir_info->fd);
	ack_info->value.iValue = smbc_lseekdir(lseekdir_info->fd, lseekdir_info->offset);
	size = sizeof(int);
	break;
    case SMBC_MSG_MKDIR:
	mkdir_info = msg;
#ifdef SMBC_USE_SHARED_MEMORY
	mkdir_info->fname = smbc_unwire_ptr(mkdir_info->h, mkdir_info->fname, SMBC_MAX_PATH);
#endif
	//printf("smbc_message_loop: MKDIR %s\n", mkdir_info->fname);
	ack_info->value.iValue = smbc_mkdir(mkdir_info->fname, mkdir_info->mode);
	size = sizeof(int);
	break;
    case SMBC_MSG_RMDIR:
	rmdir_info = msg;
#ifdef SMBC_USE_SHARED_MEMORY
	rmdir_info->durl = smbc_unwire_ptr(rmdir_info->h, rmdir_info->durl, SMBC_MAX_PATH);
#endif
	//printf("smbc_message_loop: RMDIR %s\n", rmdir_info->durl);
	ack_info->value.iValue = smbc_rmdir(rmdir_info->durl);
	size = sizeof(int);
	break;
    case SMBC_MSG_STAT:
	stat_info = msg;
#ifdef SMBC_USE_SHARED_MEMORY
	stat_info->fname = smbc_unwire_ptr(stat_info->h, stat_info->fname, SMBC_MAX_PATH);
	stat_info->st    = smbc_unwire_ptr(stat_info->h, stat_info->st, sizeof(struct stat));
#endif
	//printf("smbc_message_loop: STAT %s\n", stat_info->fname);
	ack_info->value.iValue = smbc_stat(stat_info->fname, stat_info->st);
	size = sizeof(int);
	break;
    case SMBC_MSG_FSTAT:
	fstat_info = msg;
#ifdef SMBC_USE_SHARED_MEMORY
	fstat_info->st = smbc_unwire_ptr(fstat_info->h, fstat_info->st, sizeof(struct stat));
#endif
	//printf("smbc_message_loop: FSTAT %d\n", fstat_info->fd);
	ack_info->value.iValue = smbc_fstat(fstat_info->fd, fstat_info->st);
	size = sizeof(int);
	break;
#if 0
    case SMBC_MSG_CHOWN:
	chown_info = msg;
#ifdef SMBC_USE_SHARED_MEMORY
	chown_info->url = smbc_unwire_ptr(chown_info->h, chown_info->url, SMBC_MAX_PATH);
#endif
	//printf("smbc_message_loop: CHOWN %s\n", chown_info->url);
	ack_info->value.iValue = smbc_chown(chown_info->url, chown_info->owner, chown_info->group);
	size = sizeof(int);
	break;
    case SMBC_MSG_CHMOD:
	chmod_info = msg;
#ifdef SMBC_USE_SHARED_MEMORY
	chmod_info->url = smbc_unwire_ptr(chmod_info->h, chmod_info->url, SMBC_MAX_PATH);
#endif
	//printf("smbc_message_loop: CHMOD %s\n", chmod_info->url);
	ack_info->value.iValue = smbc_chmod(chmod_info->url, chmod_info->mode);
	size = sizeof(int);
	break;
#endif
    case SMBC_MSG_REMOVEUNUSEDSERVERS:
	rus_info = msg;
	//printf("smbc_message_loop: REMOVEUNUSEDSERVERS %d\n", rus_info->msec);
	smbc_remove_unused_servers(rus_info->msec);
	size = 0;
	break;
    case SMBC_MSG_REINIT:
	//printf("smbc_message_loop: REINIT\n");
	ack_info->value.iValue = smbc_reinit();
	ack_info->value.iValue = nmbc_reinit() ? -1 : ack_info->value.iValue;
	size = sizeof(int);
	break;
    case NMBC_MSG_NAMEQUERY:
        {
            struct in_addr *ip_list;
            int count = 0, flags = 0;
            nq_info = msg;

#ifdef SMBC_USE_SHARED_MEMORY
	    nq_info->name = smbc_unwire_ptr(nq_info->h, nq_info->name, SMBC_MAX_PATH);
	    nq_info->ip_addr = smbc_unwire_ptr(nq_info->h, nq_info->ip_addr, sizeof(struct in_addr));
#endif
            ip_list = nmbc_name_query(nq_info->name, 0, TRUE, FALSE,
                                      nq_info->bcast_addr, &count, &flags,
                                      NULL);
            if(!ip_list)
                ack_info->value.iValue = -1;
            else {
                /* only use 1st returned IP (?) */
                memcpy(nq_info->ip_addr, &ip_list[0], sizeof(struct in_addr));
                ack_info->value.iValue = 0;
            }
        }
        size = sizeof(int);
        break;
#ifdef SMBC_USE_SHARED_MEMORY
    case SMBC_MSG_ALLOCSHARED:
	allocsh_info = msg;
	ack_info->value.iValue = smbc_alloc_shared(allocsh_info->size);
	size = sizeof(int);
	break;
    case SMBC_MSG_FREESHARED:
	freesh_info = msg;
	ack_info->value.iValue = smbc_free_shared(freesh_info->h);
	size = sizeof(int);
	break;
#endif
    case SMBC_MSG_GETPID:
	ack_info->value.iValue = getpid();
	size = sizeof(int);
	break;
    default:
	printf("smbc_message_loop: Unknown message type %d\n", msg->type);
	ack_info->value.iValue = -1;
	size = sizeof(int);
    }
    ack_info->_errno   = errno;
    ack_info->hdr.size = FIELD_OFFSET(SMBC_ACK_INFO, value.buffer[size]);
    ack_info->hdr.type = SMBC_MSG_ACK;
    return 1;
}

int smbc_message_loop(int fd)
{
    char buf[256];
    SMBC_MSG *msg;
    interruptible_t intr;

    chkstack_init_thread("smbc_message_loop", 65536-260);

    printf("smbc_message_loop: Connected\n");
#if 0
    if (smbc_init(smbc_auth_fn, 0) < 0) {
	printf("smbc_message_loop: smb init failed, error %d\n", errno);
	close(fd);
	return 0;
    }
#endif

    //checksum_smbclient();

    //intr.handler = INTERRUPTIBLE_ALWAYS;
    //enter_interruptible(&intr);

    msg = buf;
    do {

	//printf("smbc_message_loop: Waiting for message\n");

	if (!smbc_recv_msg(fd, msg, sizeof(buf)))
	    break;
	//printf("smbc_message_loop: Got message %d\n", msg->type);
	//sleep(1);

	if (!smbc_dispatch_msg(msg, msg))
	    break;

	//checksum_smbclient();

	//printf("smbc_message_loop: Response %d\n", ack_info->value.iValue);
	if (!smbc_send_msg(fd, msg))
	    break;

    } while (1);

    printf("smbc_message_loop: exiting\n");
#ifdef SMBC_USE_SHARED_MEMORY
    smbc_free_shared_owner(pthread_self());
#endif

    //leave_interruptible(&intr);
    close(fd);
    return 0;
}

int smbc_send_msg(int fd, SMBC_MSG *msg)
{
    int i, j;

    //printf("smbc_send_msg: sending %d bytes\n", msg->size);
    for (i = 0; i < msg->size; i += j) {

	j = send(fd, ((char *)msg)+i, msg->size - i, 0);
	if (j < 0) {
	    printf("smbc_send_msg: Failed to send %d, error %d\n", j, errno);
	    return 0;
	}
    }
    //printf("smbc_send_msg: %d bytes sent\n", i);
    return i;
}

int smbc_recv_msg(int fd, SMBC_MSG *msg, int size)
{
    int i, j;

    for (i = 0; i < sizeof(SMBC_MSG); i += j) {

	j = recv(fd, ((char *)msg) + i, sizeof(SMBC_MSG) - i, 0);
	if (j <= 0) {
	    printf("smbc_recv_msg: Failed to receive %d, error %d\n", j, errno);
	    return 0;
	}
    }

    if (msg->size > size) {
	printf("smbc_recv_msg: Message too large %d\n", msg->size);
	return 0;
    }

    for (i; i < msg->size; i += j) {

	j = recv(fd, ((char *)msg) + i, msg->size - i, 0);
	if (j <= 0) {
	    printf("smbc_recv_msg: Failed to receive %d, error %d\n", j, errno);
	    return 0;
	}
    }
    return i;
}

pthread_t smbc_create_thread(void *(*thread_proc)(void *pv), void *pv)
{
    pthread_t id = 0;
    pthread_attr_t attr;

    pthread_attr_init(&attr);
    pthread_attr_setscope(&attr, 0);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    pthread_attr_setstacksize(&attr, 65536);
    pthread_create(&id, &attr, thread_proc, pv);
    pthread_attr_destroy(&attr);
    return id;
}

int smbc_listen_thread(void)
{
    int s, a, addrlen;
    struct sockaddr_in addr;

    s = socket(AF_INET, SOCK_STREAM, 0);
    if (s < 0) {
	printf("smbc_listen_thread: Couldn't create socket (error %d)\n", errno);
	return 0;
    }

    memset(&addr, 0, sizeof(addr));
    addr.sin_family      = AF_INET;
    addr.sin_port        = htons(SMBC_REMOTE_PORT);
    addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
    if (bind(s, &addr, sizeof(addr)) < 0) {
	printf("smbc_listen_thread: Couldn't bind to port (error %d)\n", errno);
	return 0;
    }

    if (listen(s, 5) < 0) {
	printf("smbc_listen_thread: Couldn't listen (error %d)\n", errno);
	return 0;
    }

    printf("smbc_listen_thread: Started\n");
    do {

	addrlen = sizeof(addr);
	a = accept(s, &addr, &addrlen);
	if (a >= 0) {
	    smbc_create_thread(smbc_message_loop, a);
	    //smbc_message_loop(a);
	    //close(a);
	}
    } while (1);
    return 0;
}

#define NOP()   	__asm("andeq\tr0,r0,r0")
#define NOP10() 	{ NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); NOP(); }
#define NOP100()	{ NOP10(); NOP10(); NOP10(); NOP10(); NOP10(); NOP10(); NOP10(); NOP10(); NOP10(); NOP10(); }
#define NOP1000()       { NOP100(); NOP100(); NOP100(); NOP100(); NOP100(); NOP100(); NOP100(); NOP100(); NOP100(); NOP100(); }
#define NOP10000()	{ NOP1000(); NOP1000(); NOP1000(); NOP1000(); NOP1000(); NOP1000(); NOP1000(); NOP1000(); NOP1000(); NOP1000(); }

//char filldata[9995] = {-1};

int main(int argc, char *argv[])
{
#ifdef ARM
    __morecore = my_morecore;
#endif

    if (smbc_init(smbc_auth_fn, 0) < 0) {
	printf("smbclient: smb init failed, error %d\n", errno);
	return 1;
    }

    pthread_mutex_init(&g_smbc_mutex, NULL);

    init_interruptible();

    //smbc_create_thread(smbc_redir_thread, NULL);
    //checksum_smbclient();
    //malloc(1013);
    return smbc_listen_thread();
}
