/*
 * arch/arm/mach-em86xx/arch.c
 *
 * Copyright (C) 2003-2004 Sigma Designs, Inc
 *
 * by Ho Lee 01/27/2003
 */
#include <linux/config.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/init.h>

#include <asm/hardware.h>
#include <asm/irq.h>
#include <asm/setup.h>
#include <asm/mach-types.h>

#include <asm/mach/arch.h>
#include <asm/arch/em86xxapi.h>
#include <asm/arch/param.h>

/* The size of memory used by kernel */
unsigned long em86xx_kmemsize = DRAM_SIZE;

#ifdef CONFIG_SD_USE_BOOTLOADER_MEMCFG
memcfg_t *em86xx_memcfg_ptr = (memcfg_t *)MEMCFG_START;
#else /* ! CONFIG_SD_USE_BOOTLOADER_MEMCFG */
static memcfg_t em86xx_memcfg = {
	signature: MEMCFG_SIGNATURE,
	dram0_size: DRAM_SIZE,
#ifdef CONFIG_NUMA_EM86XX
	dram1_size: DRAM_SIZE_2,
#else
	dram1_size: 0,
#endif
	dram2_size: 0,
	dram0_removable_topreserved: 0,
	dram1_removable_topreserved: 0,
	dram2_removable_topreserved: 0,
	dram0_fixed_topreserved: 0,
	dram1_fixed_topreserved: 0,
	dram2_fixed_topreserved: 0,
	kernel_end: 0x00080000,	/* By default reserve first 512KB in DRAM0 */
	checksum: 0,	/* Not a valid checksum */
	reserved: {0, 0, 0, 0},
};
memcfg_t *em86xx_memcfg_ptr = (memcfg_t *)&em86xx_memcfg;
#endif /* CONFIG_SD_USE_BOOTLOADER_MEMCFG */

extern void em86xx_init_irq(void);

static void __init em86xx_fixup(struct machine_desc *desc, struct param_struct *unusd,
    char **cmdline, struct meminfo *mi)
{
    unsigned int revid = __raw_readl(REG_BASE_HOST + PCI_devcfg_reg1) & 0x0f;

#if defined(CONFIG_ARCH_MAMBO)
    printk("Mambo Rev %c (kernel supports Rev %c)\n", 
        'A' + revid - 1, 'A' + CONFIG_ARCH_MAMBO_REV - 1);
#elif defined(CONFIG_ARCH_TANGO10)
    // No Rev. B
    if (revid >= 2)
        ++revid;
    printk("Tango10 Rev %c (kernel supports Rev %c)\n", 
        'A' + revid - 1, 'A' + CONFIG_ARCH_TANGO10_REV - 1);
#elif defined(CONFIG_ARCH_TANGO15)
    printk("Tango15 Rev %c (kernel supports Rev %c)\n", 
        'A' + revid - 1, 'A' + CONFIG_ARCH_TANGO15_REV - 1);
#endif
    printk("Board name is %s\n", CONFIG_ARCH_EM86XX_BOARD);
    
    em86xx_sbox_init();
}

MACHINE_START(EM86XX, "EM86XX")
    MAINTAINER("Craig Qu - Sigma Designs, Inc")
    BOOT_MEM(PHYS_OFFSET, PHYS_OFFSET, PHYS_OFFSET) // phys_ram, phys_io, virt_io
                                                    // phys_io, virt_io have no meaning
                                                    // but if virt_io is less than 0x00040000, kernel doesn't boot
                                                    // (refer __lookup_architecture_type() at head-armv.S
                                                    // r7 = (virt_io >> 18) 
    INITIRQ(em86xx_init_irq)
    FIXUP(em86xx_fixup)
    BOOT_PARAMS(DRAM_BASE + 0x200)
MACHINE_END
