/*
 * Make Kernel with file header
 *
 * Copyright (c) SkipJam Corp
 *
 */
 
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <aio.h>
#include <pthread.h>
#include <signal.h>
#include <time.h>
#include <termios.h>

#define MAX_HEADER_BLOCKS           4
#define MAX_FILENAME_LENGTH         ((MAX_HEADER_BLOCKS-1)*16) //uses 3 blocks of 16 total for header

#pragma pack(1)

static __inline__ unsigned int swapl(unsigned int x)
{
    return ((unsigned long int)((((unsigned long int)(x) & 0x000000ffU) << 24) |
        (((unsigned long int)(x) & 0x0000ff00U) <<  8) |
        (((unsigned long int)(x) & 0x00ff0000U) >>  8) |
        (((unsigned long int)(x) & 0xff000000U) >> 24)));
}

typedef struct FSHEADERSTRUCT {
    unsigned int next;
    unsigned int info;
    unsigned int size;
    unsigned int checksum;
    char filename[MAX_FILENAME_LENGTH];              //max file name length supported
} FSHEADER;
        
FSHEADER fsheader;

#define MAX_KERNEL_SIZE             (0x1a0000 - sizeof(FSHEADER))

char buff[512];

main(int argc, char *argv[])
{
    unsigned int infil, fsize, hdrsize, rdsize, i;
	unsigned int chksum = 0;
    char *val,*fname;
    
    if(argc < 2) {
        fprintf(stderr,"usage: %s infile\n",argv[0]);
        exit(1);
    }
    
    if( (infil = open(argv[1],O_RDONLY)) < 0) {
        fprintf(stderr,"%s: Unable to open Kernel file %s\n",argv[0],argv[1]);
        exit(1);
    }
    if( !(fsize = lseek(infil, 0L, SEEK_END)) ) {
        fprintf(stderr,"%s: Kernel file 0 lenghth\n",argv[0]);
        exit(1);
    }
    if(fsize > MAX_KERNEL_SIZE) {
        fprintf(stderr,"%s: Kernel size 0x%08x exceeds maximum 0x%08x\n",fsize, MAX_KERNEL_SIZE);
        exit(1);
    }
    
    fsheader.next = 0;
    fsheader.info = 0;
    fsheader.checksum = 0;
    fsheader.size = swapl(fsize);
    fname = argv[1] + strlen(argv[1]) - 1;
    while( fname && fname >= argv[1] && *fname) {
        if(*(fname--) == '\\')
            break;
    }
    ++fname;
    if(*fname == '\\')
        ++fname;
    if(strlen(fname) > MAX_FILENAME_LENGTH-1) {
        fprintf(stderr,"%s: File name %s exceeds max of 19 characters\n",argv[0],fname);
        exit(1);
    }
    
    strcpy(fsheader.filename, fname);
    
    val = (unsigned char *) &fsheader;
    for(i = 0; i < 32; ++i, ++val)
        chksum += (*val & 0xff);
    fsheader.checksum = chksum;
    
    hdrsize = sizeof(FSHEADER) - (16 * ((MAX_HEADER_BLOCKS-2) - (strlen(fname) / 16)));
                        
    fprintf(stderr,"%s info: File name is %s\n",argv[0], fsheader.filename);
	fprintf(stderr,"%s info: Header checksum is 0x%08x\n",argv[0], chksum);
    fprintf(stderr,"%s info: File size is 0x%08x / %d\n",argv[0], fsize, fsize);
    fprintf(stderr,"%s info: Header size is 0x%08x / %d\n",argv[0], hdrsize, hdrsize);
    
    if(write(1, (void *) &fsheader, hdrsize) != hdrsize) {
        fprintf(stderr, "%s: write error\n",argv[0]);
        exit (1);
    }
    
    lseek(infil, 0L, SEEK_SET);
    while((rdsize = read(infil, buff, sizeof(buff))) != 0)
        write(1, buff, rdsize);
    
    close(infil);
    close(1);
    exit(0);
}
